use crate::Imgui;

use bytemuck::{
    Pod,
    Zeroable,
};

use std::time::Instant;

use imgui::Condition;

use winit::window::Window;
use winit::dpi::PhysicalSize;
use winit::event::WindowEvent;

use wgpu::util::{
    DeviceExt,
    BufferInitDescriptor,
};
use wgpu::{
    Face,
    Color,
    Queue,
    Buffer,
    Device,
    LoadOp,
    Surface,
    Backends,
    Instance,
    FrontFace,
    BlendState,
    Operations,
    ColorWrites,
    PolygonMode,
    VertexState,
    BufferUsages,
    VertexFormat,
    BufferAddress,
    FragmentState,
    TextureUsages,
    PrimitiveState,
    RenderPipeline,
    VertexStepMode,
    PowerPreference,
    VertexAttribute,
    ColorTargetState,
    DeviceDescriptor,
    MultisampleState,
    PrimitiveTopology,
    InstanceDescriptor,
    VertexBufferLayout,
    RenderPassDescriptor,
    SurfaceConfiguration,
    RequestAdapterOptions,
    TextureViewDescriptor,
    CommandEncoderDescriptor,
    PipelineLayoutDescriptor,
    RenderPipelineDescriptor,
    RenderPassColorAttachment,
};

#[repr(C)]
#[derive(Clone, Copy, Debug, Pod, Zeroable)]
struct Vertex {
    pos: [f32; 3],
    color: [f32; 3]
}

impl Vertex {
    fn desc() -> VertexBufferLayout<'static> {
        VertexBufferLayout {
            array_stride: std::mem::size_of::<Vertex>() as BufferAddress,
            step_mode: VertexStepMode::Vertex,
            attributes: &[
                VertexAttribute {
                    offset: 0,
                    shader_location: 0,
                    format: VertexFormat::Float32x3,
                },
                VertexAttribute {
                    offset: std::mem::size_of::<[f32; 3]>() as BufferAddress,
                    shader_location: 1,
                    format: VertexFormat::Float32x3,
                }
            ]
        }
    }
}

const VERTICES: &[Vertex] = &[
    Vertex { pos: [0.0, 1.0, 0.0], color: [1.0, 0.0, 0.0] },
    Vertex { pos: [-1.0, -1.0, 0.0], color: [0.0, 1.0, 0.0] },
    Vertex { pos: [1.0, -1.0, 0.0], color: [0.0, 0.0, 1.0] },
];

pub struct State {
    pub queue: Queue,
    pub buffer: Buffer,
    pub device: Device,
    pub window: Window,
    pub surface: Surface,
    pub last_frame: Instant,
    pub size: PhysicalSize<u32>,
    pub config: SurfaceConfiguration,
    pub render_pipeline: RenderPipeline,
}

impl State {
    pub async fn new(window: Window) -> Self {
        let instance = Instance::new(InstanceDescriptor {
            backends: Backends::all(),
            ..Default::default()
        });

        let surface = unsafe { instance.create_surface(&window) }
            .expect("Couldn't create surface");

        let adapter = instance
            .request_adapter(&RequestAdapterOptions {
                power_preference: PowerPreference::HighPerformance,
                compatible_surface: Some(&surface),
                force_fallback_adapter: false,
            })
            .await
            .expect("No supported adapters found");


        let (device, queue) = adapter
            .request_device(&DeviceDescriptor::default(), None)
            .await
            .expect("Couldn't get logic device");

        let size = window.inner_size();
        let surface_caps = surface.get_capabilities(&adapter);
        let surface_format = surface_caps.formats
            .iter()
            .copied()
            .find(|f| f.is_srgb())
            .unwrap_or(surface_caps.formats[0]);

        let config = SurfaceConfiguration {
            width: size.width,
            height: size.height,
            view_formats: vec![],
            format: surface_format,
            usage: TextureUsages::RENDER_ATTACHMENT,
            alpha_mode: surface_caps.alpha_modes[0],
            present_mode: surface_caps.present_modes[0],
        };

        surface.configure(&device, &config);

        let last_frame = Instant::now();

        let shader = device.create_shader_module(wgpu::include_wgsl!("shader.wgsl"));

        let render_pipeline_layout = device.create_pipeline_layout(&PipelineLayoutDescriptor {
            label: Some("Render Pipeline Layout"),
            bind_group_layouts: &[],
            push_constant_ranges: &[],
        });

        let render_pipeline = device.create_render_pipeline(&RenderPipelineDescriptor {
            label: Some("Render Pipeline"),
            layout: Some(&render_pipeline_layout),
            vertex: VertexState {
                module: &shader,
                entry_point: "vs_main",
                buffers: &[Vertex::desc()],
            },
            fragment: Some(FragmentState {
                module: &shader,
                entry_point: "fs_main",
                targets: &[Some(ColorTargetState {
                    format: config.format,
                    blend: Some(BlendState::REPLACE),
                    write_mask: ColorWrites::ALL,
                })],
            }),
            primitive: PrimitiveState {
                topology: PrimitiveTopology::TriangleList,
                strip_index_format: None,
                front_face: FrontFace::Ccw,
                cull_mode: Some(Face::Back),
                polygon_mode: PolygonMode::Fill,
                unclipped_depth: false,
                conservative: false,
            },
            depth_stencil: None,
            multisample: MultisampleState {
                count: 1,
                mask: !0,
                alpha_to_coverage_enabled: false,
            },
            multiview: None,
        });

        let buffer = device.create_buffer_init(&BufferInitDescriptor {
            label: Some("Vertex Buffer"),
            contents: bytemuck::cast_slice(VERTICES),
            usage: BufferUsages::VERTEX,
        });

        State {
            size,
            queue,
            buffer,
            config,
            device,
            window,
            surface,
            last_frame,
            render_pipeline,
        }
    }

    pub fn window(&self) -> &Window {
        &self.window
    }

    pub fn resize(&mut self, new_size: PhysicalSize<u32>) {
        if new_size.height <= 0 || new_size.width <= 0 {
            return;
        }

        self.size = new_size;
        self.config.width = new_size.width;
        self.config.height = new_size.height;
        self.surface.configure(&self.device, &self.config);
    }

    pub fn input(&mut self, _event: &WindowEvent) -> bool {
        false
    }

    pub fn render(&mut self, imgui: &mut Imgui) {
        let mut profile = Instant::now();
        let now = Instant::now();
        imgui.context
            .io_mut()
            .update_delta_time(now-self.last_frame);
        self.last_frame = now;

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 1 took: {:?}ms", delta);
        }

        let frame = match self.surface.get_current_texture() {
            Ok(frame) => frame,
            Err(wgpu::SurfaceError::Lost) => {
                self.resize(self.size);
                return;
            },
            Err(e) => {
                println!("pauuuu {:?}", e);
                return;
            },
        };

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 2 took: {:?}ms", delta);
        }

        imgui.platform
            .prepare_frame(imgui.context.io_mut(), self.window())
            .expect("Failed to prepare frame");

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point3 took: {:?}ms", delta);
        }

        let ui = imgui.context.frame();
        {
            let window = ui.window("Hello world");
            window
                .size([300.0, 100.0], Condition::FirstUseEver)
                .build(|| {
                    ui.text("Hello world!");
                    ui.text("This...is...imgui-rs on WGPU!");
                    ui.separator();
                    let mouse_pos = ui.io().mouse_pos;
                    ui.text(format!(
                        "Mouse Position: ({:.1},{:.1})",
                        mouse_pos[0], mouse_pos[1]
                    ));
                });
        }

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 4 took: {:?}ms", delta);
        }

        if imgui.last_cursor != ui.mouse_cursor() {
            imgui.last_cursor = ui.mouse_cursor();
            imgui.platform.prepare_render(ui, &self.window());
        }

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 5 took: {:?}ms", delta);
        }

        let view = frame.texture.create_view(&TextureViewDescriptor::default());
        let mut encoder = self.device
            .create_command_encoder(&CommandEncoderDescriptor {
                label: None
            });

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 6 took: {:?}ms", delta);
        }

        {
            let render_pass = RenderPassDescriptor {
                label: Some("Render Pass"),
                color_attachments: &[Some(RenderPassColorAttachment {
                    view: &view,
                    resolve_target: None,
                    ops: Operations {
                        load: LoadOp::Clear(Color {
                            r: 0.1,
                            g: 0.2,
                            b: 0.3,
                            a: 1.0,
                        }),
                        store: true,
                    },
                })],
                depth_stencil_attachment: None,
            };

            let mut render_pass = encoder.begin_render_pass(&render_pass);

            render_pass.set_pipeline(&self.render_pipeline);
            render_pass.set_vertex_buffer(0, self.buffer.slice(..));
            render_pass.draw(0..(VERTICES.len() as u32), 0..1);

            imgui.renderer.render(
                imgui.context.render(),
                &self.queue,
                &self.device,
                &mut render_pass
            ).expect("Rendering failed");
        }

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 7 took: {:?}ms", delta);
        }

        self.queue.submit(Some(encoder.finish()));

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 8 took: {:?}ms", delta);
        }

        frame.present();

        {
            let n = Instant::now();
            let delta = (n-profile).as_secs_f64()*1000.0;
            profile = n;
            println!("Point 9 took: {:?}ms", delta);
        }
    }
}
